import io.github.jaredmdobson.concentus.OpusDecoder
import io.github.jaredmdobson.concentus.OpusEncoder
import java.util.concurrent.atomic.AtomicInteger
import javax.sound.sampled.AudioFormat
import kotlin.math.max
import kotlin.math.min

object AudioConvert {
    fun getAudioFormat(sampleRate: Int, channels: Int): AudioFormat {
        return AudioFormat(
            AudioFormat.Encoding.PCM_SIGNED,
            sampleRate.toFloat(),
            16,
            channels,
            channels * 2,
            sampleRate.toFloat(),
            false
        )
    }

    /**
     * 将 float[] 转换为 short 的 byte[]。
     *
     * @param input  输入的 float[] 数据
     * @param output 转换后的 byte[] 数据
     */
    fun convertFloatToShortByte(input: FloatArray, output: ByteArray) {
        require(input.size * 2 == output.size) { "output byte array length must be twice the length of the input float array." }

        for (i in input.indices) {
            // 将 float 值限制在 -1.0 到 1.0 之间
            val sample = min(max(input[i].toDouble(), -1.0), 1.0).toFloat()

            // 将 float 值转换为 short
            val shortSample = (sample * 32767).toInt().toShort()

            // 将 short 值转换为两个字节，并存储在 lineBuffer 中
            output[i * 2] = (shortSample.toInt() and 0xFF).toByte() // 低位字节
            output[i * 2 + 1] = ((shortSample.toInt() shr 8) and 0xFF).toByte() // 高位字节
        }
    }

    /**
     * 将 short 的 byte[] 转换为 float[]。
     *
     * @param input  输入的 byte[] 数据（每两个字节表示一个 short）
     * @param output 转换后的 float[] 数据
     */
    fun convertShortByteToFloat(input: ByteArray, output: FloatArray) {
        require(input.size == output.size * 2) { "Input byte array length must be twice the length of the output float array." }

        for (i in output.indices) {
            // 将两个字节转换为 short
            val low = input[i * 2].toInt() and 0xFF
            val high = input[i * 2 + 1].toInt() and 0xFF
            val shortSample = ((high shl 8) or low).toShort()

            // 将 short 值转换为 float
            output[i] = shortSample / 32768.0f
        }
    }

    fun bytes2shorts(input: ByteArray, output: ShortArray) {
        for (c in output.indices) {
            val low = input[(c * 2)].toInt() and 0xFF
            val high = input[(c * 2) + 1].toInt() shl 8
            output[c] = (low or high).toShort()
        }
    }

    fun shorts2bytes(input: ShortArray, output: ByteArray) {
        for (c in input.indices) {
            output[c * 2] = (input[c].toInt() and 0xFF).toByte()
            output[c * 2 + 1] = ((input[c].toInt() shr 8) and 0xFF).toByte()
        }
    }

}


// 下面是统计用的
val compressRateOpus = AtomicInteger()
val compressRatePcm = AtomicInteger()


val encodeOpusBuffer = mutableMapOf<OpusEncoder, ByteArray>()
val encodePcmShortBuffer = mutableMapOf<OpusEncoder, ShortArray>()

// 1275是opus单帧包最大值
fun OpusEncoder.encode(pcm: ByteArray): DataPack {
    if (RAW_PCM_PASS) {
        return DataPack(System.currentTimeMillis(), pcm.copyOf())
    }
    val opusBuffer = encodeOpusBuffer[this] ?: ByteArray(1275).also { encodeOpusBuffer[this] = it }
    val shortBuffer = encodePcmShortBuffer[this] ?: ShortArray(pcm.size / 2).also { encodePcmShortBuffer[this] = it }
    AudioConvert.bytes2shorts(pcm, shortBuffer)
    val bytesEncoded = encode(shortBuffer, 0, PACKET_SAMPLES, opusBuffer, 0, opusBuffer.size)
    compressRatePcm.addAndGet(pcm.size)
    compressRateOpus.addAndGet(bytesEncoded)
    return DataPack(System.currentTimeMillis(), opusBuffer.sliceArray(0..<bytesEncoded))
}

val decodePcmShortBuffer = mutableMapOf<OpusDecoder, ShortArray>()
fun OpusDecoder.decode(pack: DataPack, pcm: ByteArray) {
    if (RAW_PCM_PASS) {
        pack.opus.copyInto(pcm)
        return
    }
    val shortBuffer = decodePcmShortBuffer[this] ?: ShortArray(pcm.size / 2).also { decodePcmShortBuffer[this] = it }
    decode(pack.opus, 0, pack.opus.size, shortBuffer, 0, PACKET_SAMPLES, false)
    AudioConvert.shorts2bytes(shortBuffer, pcm)
}

fun <T> T.loge() = also { println(this) }